@@ -3,27 +3,28 @@
 Monitor & Cull idle single-user servers and users
 """
 
+import asyncio
 import ssl
 import json
 import os
 from datetime import datetime
 from datetime import timezone
+from distutils.version import LooseVersion as V
 from functools import partial
+from textwrap import dedent
 
-try:
-    from urllib.parse import quote
-except ImportError:
-    from urllib import quote
+from urllib.parse import quote
 
 import dateutil.parser
 
-from tornado.gen import coroutine, multi
-from tornado.locks import Semaphore
 from tornado.log import app_log
 from tornado.httpclient import AsyncHTTPClient, HTTPRequest
+from tornado.httputil import url_concat
 from tornado.ioloop import IOLoop, PeriodicCallback
 from tornado.options import define, options, parse_command_line
 
+STATE_FILTER_MIN_VERSION = V("1.3.0")
+
 
 def parse_date(date_string):
     """Parse a timestamp
@@ -69,8 +70,7 @@ def make_ssl_context(keyfile, certfile, cafile=None, verify=True, check_hostname
     return ssl_context
 
 
-@coroutine
-def cull_idle(
+async def cull_idle(
     url,
     api_token,
     inactive_limit,
@@ -80,15 +80,19 @@ def cull_idle(
     concurrency=10,
     ssl_enabled=False,
     internal_certs_location="",
+    cull_admin_users=True,
+    api_page_size=0,
 ):
     """Shutdown idle single-user servers
 
     If cull_users, inactive *users* will be deleted as well.
     """
-    auth_header = {"Authorization": "token %s" % api_token}
-    req = HTTPRequest(url=url + "/users", headers=auth_header)
-    now = datetime.now(timezone.utc)
-
+    defaults = {
+        # GET /users may be slow if there are thousands of users and we
+        # don't do any server side filtering so default request timeouts
+        # to 60 seconds rather than tornado's 20 second default.
+        "request_timeout": os.environ.get("JUPYTERHUB_REQUEST_TIMEOUT", 60)
+    }
     if ssl_enabled:
         ssl_context = make_ssl_context(
             f"{internal_certs_location}/hub-internal/hub-internal.key",
@@ -98,31 +102,73 @@ def cull_idle(
 
         app_log.debug("ssl_enabled is Enabled: %s", ssl_enabled)
         app_log.debug("internal_certs_location is %s", internal_certs_location)
-        AsyncHTTPClient.configure(None, defaults={"ssl_options": ssl_context})
+        defaults["ssl_options"] = ssl_context
 
+    AsyncHTTPClient.configure(None, defaults=defaults)
     client = AsyncHTTPClient()
 
     if concurrency:
-        semaphore = Semaphore(concurrency)
+        semaphore = asyncio.Semaphore(concurrency)
 
-        @coroutine
-        def fetch(req):
+        async def fetch(req):
             """client.fetch wrapped in a semaphore to limit concurrency"""
-            yield semaphore.acquire()
+            await semaphore.acquire()
             try:
-                return (yield client.fetch(req))
+                return await client.fetch(req)
             finally:
-                yield semaphore.release()
+                semaphore.release()
 
     else:
         fetch = client.fetch
 
-    resp = yield fetch(req)
-    users = json.loads(resp.body.decode("utf8", "replace"))
-    futures = []
+    async def fetch_paginated(req):
+        """Make a paginated API request
 
-    @coroutine
-    def handle_server(user, server_name, server, max_age, inactive_limit):
+        async generator, yields all items from a list endpoint
+        """
+        req.headers["Accept"] = "application/jupyterhub-pagination+json"
+        url = req.url
+        resp_future = asyncio.ensure_future(fetch(req))
+        page_no = 1
+        item_count = 0
+        while resp_future is not None:
+            response = await resp_future
+            resp_future = None
+            resp_model = json.loads(response.body.decode("utf8", "replace"))
+
+            if isinstance(resp_model, list):
+                # handle pre-2.0 response, no pagination
+                items = resp_model
+            else:
+                # paginated response
+                items = resp_model["items"]
+
+                next_info = resp_model["_pagination"]["next"]
+                if next_info:
+                    page_no += 1
+                    app_log.info(f"Fetching page {page_no} {next_info['url']}")
+                    # submit next request
+                    req.url = next_info["url"]
+                    resp_future = asyncio.ensure_future(fetch(req))
+
+            for item in items:
+                item_count += 1
+                yield item
+
+        app_log.debug(f"Fetched {item_count} items from {url} in {page_no} pages")
+
+    # Starting with jupyterhub 1.3.0 the users can be filtered in the server
+    # using the `state` filter parameter. "ready" means all users who have any
+    # ready servers (running, not pending).
+    auth_header = {"Authorization": "token %s" % api_token}
+    resp = await fetch(HTTPRequest(url=url + "/info", headers=auth_header))
+
+    info = json.loads(resp.body.decode("utf8", "replace"))
+    state_filter = V(info["version"]) >= STATE_FILTER_MIN_VERSION
+
+    now = datetime.now(timezone.utc)
+
+    async def handle_server(user, server_name, server, max_age, inactive_limit):
         """Handle (maybe) culling a single server
 
         "server" is the entire server model from the API.
@@ -236,15 +282,14 @@ def cull_idle(
             body=body,
             allow_nonstandard_methods=True,
         )
-        resp = yield fetch(req)
+        resp = await fetch(req)
         if resp.code == 202:
             app_log.warning("Server %s is slow to stop", log_name)
             # return False to prevent culling user with pending shutdowns
             return False
         return True
 
-    @coroutine
-    def handle_user(user):
+    async def handle_user(user):
         """Handle one user.
 
         Create a list of their servers, and async exec them.  Wait for
@@ -273,7 +318,11 @@ def cull_idle(
             handle_server(user, server_name, server, max_age, inactive_limit)
             for server_name, server in servers.items()
         ]
-        results = yield multi(server_futures)
+        if server_futures:
+            results = await asyncio.gather(*server_futures)
+        else:
+            results = []
+
         if not cull_users:
             return
         # some servers are still running, cannot cull users
@@ -303,9 +352,12 @@ def cull_idle(
             # which introduces the 'created' field which is never None
             inactive = age
 
+        user_is_admin = user["admin"]
+
         should_cull = (
             inactive is not None and inactive.total_seconds() >= inactive_limit
-        )
+        ) and (cull_admin_users or not user_is_admin)
+
         if should_cull:
             app_log.info("Culling user %s (inactive for %s)", user["name"], inactive)
 
@@ -334,15 +386,50 @@ def cull_idle(
         req = HTTPRequest(
             url=url + "/users/%s" % user["name"], method="DELETE", headers=auth_header
         )
-        yield fetch(req)
+        await fetch(req)
         return True
 
-    for user in users:
+    futures = []
+
+    params = {}
+    if api_page_size:
+        params["limit"] = str(api_page_size)
+
+    users_url = url + "/users"
+
+    # If we filter users by state=ready then we do not get back any which
+    # are inactive, so if we're also culling users get the set of users which
+    # are inactive and see if they should be culled as well.
+    if state_filter and cull_users:
+        inactive_params = {"state": "inactive"}
+        inactive_params.update(params)
+        req = HTTPRequest(url_concat(users_url, inactive_params), headers=auth_header)
+        n_idle = 0
+        async for user in fetch_paginated(req):
+            n_idle += 1
+            futures.append((user["name"], handle_user(user)))
+        app_log.debug(f"Got {n_idle} users with inactive servers")
+
+    if state_filter:
+        params["state"] = "ready"
+
+    req = HTTPRequest(
+        url=url_concat(users_url, params),
+        headers=auth_header,
+    )
+
+    n_users = 0
+    async for user in fetch_paginated(req):
+        n_users += 1
         futures.append((user["name"], handle_user(user)))
 
+    app_log.debug(
+        "Got %d users%s", n_users, (" with ready servers" if state_filter else "")
+    )
+
     for (name, f) in futures:
         try:
-            result = yield f
+            result = await f
         except Exception:
             app_log.exception("Error processing %s", name)
         else:
@@ -354,56 +441,120 @@ def main():
     define(
         "url",
         default=os.environ.get("JUPYTERHUB_API_URL"),
-        help="The JupyterHub API URL",
+        help=dedent(
+            """
+            The JupyterHub API URL.
+            """
+        ).strip(),
+    )
+    define(
+        "timeout",
+        type=int,
+        default=600,
+        help=dedent(
+            """
+            The idle timeout (in seconds).
+            """
+        ).strip(),
     )
-    define("timeout", type=int, default=600, help="The idle timeout (in seconds)")
     define(
         "cull_every",
         type=int,
         default=0,
-        help="The interval (in seconds) for checking for idle servers to cull",
+        help=dedent(
+            """
+            The interval (in seconds) for checking for idle servers to cull.
+            """
+        ).strip(),
     )
     define(
         "max_age",
         type=int,
         default=0,
-        help="The maximum age (in seconds) of servers that should be culled even if they are active",
+        help=dedent(
+            """
+            The maximum age (in seconds) of servers that should be culled even if they are active.
+            """
+        ).strip(),
     )
     define(
         "cull_users",
         type=bool,
         default=False,
-        help="""Cull users in addition to servers.
-                This is for use in temporary-user cases such as tmpnb.""",
+        help=dedent(
+            """
+            Cull users in addition to servers.
+
+            This is for use in temporary-user cases such as tmpnb.
+            """
+        ).strip(),
     )
     define(
         "remove_named_servers",
         default=False,
         type=bool,
-        help="""Remove named servers in addition to stopping them.
-            This is useful for a BinderHub that uses authentication and named servers.""",
+        help=dedent(
+            """
+            Remove named servers in addition to stopping them.
+
+            This is useful for a BinderHub that uses authentication and named servers.
+            """
+        ).strip(),
     )
     define(
         "concurrency",
         type=int,
         default=10,
-        help="""Limit the number of concurrent requests made to the Hub.
+        help=dedent(
+            """
+            Limit the number of concurrent requests made to the Hub.
 
-                Deleting a lot of users at the same time can slow down the Hub,
-                so limit the number of API requests we have outstanding at any given time.
-                """,
+            Deleting a lot of users at the same time can slow down the Hub,
+            so limit the number of API requests we have outstanding at any given time.
+            """
+        ).strip(),
     )
     define(
         "ssl_enabled",
         type=bool,
         default=False,
-        help="Whether the Jupyter API endpoint has TLS enabled",
+        help=dedent(
+            """
+            Whether the Jupyter API endpoint has TLS enabled.
+            """
+        ).strip(),
     )
     define(
         "internal_certs_location",
         type=str,
         default="internal-ssl",
-        help="The location of generated internal-ssl certificates (only needed if ssl_enabled=True)",
+        help=dedent(
+            """
+            The location of generated internal-ssl certificates (only needed with --ssl-enabled=true).
+            """
+        ).strip(),
+    )
+    define(
+        "cull_admin_users",
+        type=bool,
+        default=True,
+        help=dedent(
+            """
+            Whether admin users should be culled (only if --cull-users=true).
+            """
+        ).strip(),
+    )
+    define(
+        "api_page_size",
+        type=int,
+        default=0,
+        help=dedent(
+            """
+            Number of users to request per page,
+            when using JupyterHub 2.0's paginated user list API.
+            Default: user the server-side default configured page size.
+            """
+        ).strip(),
     )
 
     parse_command_line()
@@ -432,6 +583,8 @@ def main():
         concurrency=options.concurrency,
         ssl_enabled=options.ssl_enabled,
         internal_certs_location=options.internal_certs_location,
+        cull_admin_users=options.cull_admin_users,
+        api_page_size=options.api_page_size,
     )
     # schedule first cull immediately
     # because PeriodicCallback doesn't start until the end of the first interval
